package ru.landgrafhomyak.db.serdha0.api.module

import ru.landgrafhomyak.db.serdha0.api.table.Table
import ru.landgrafhomyak.db.serdha0.api.table.TableConstructor

/**
 * Scope methods for upgrading [module template][ModuleTemplate].
 *
 * @see ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate
 * @see ModuleTemplate.ModuleUpgrade.upgradeSchema
 */
public interface UpgradeModuleScope<omUE : Any> : CreateModuleScope {
	public val oldModule: omUE

	/**
	 * Functional interface for updating table.
	 *
	 * Can be implemented like a companion object that just calls [tnUE]'s constructor with the same or similar signature.
	 *
	 * @param tnUE User's type with table's descriptors from the previous table version.
	 * @param toUE User's type with descriptors of upgraded table.
	 *
	 * @see UpgradeModuleScope.upgradeTable
	 * @see UpgradeModuleScope.renameTable
	 * @see UpgradeModuleScope.upgradeAndRenameTable
	 * @see UpgradeModuleScope.CreateTable
	 */
	public interface TableUpgrade<tnUE : Any, toUE : Any> {
		/**
		 * Scope method that uses [upgrader] to update table schema.
		 *
		 * @param oldTable Descriptor to the previous version of table from which can be got user's extension with old descriptors.
		 * @param upgrader Object with descriptor providers and factories.
		 * @return User's object with upgraded descriptors for future access.
		 *
		 * @see UpgradeModuleScope.upgradeTable
		 * @see UpgradeModuleScope.renameTable
		 * @see UpgradeModuleScope.upgradeAndRenameTable
		 * @see UpgradeModuleScope.CreateTable
		 */
		public fun upgradeTable(oldTable: Table<toUE, *>, upgrade: TableUpgrade<tnUE, toUE>): tnUE
	}


	/**
	 * Marks that table wouldn't be changed in this module version.
	 *
	 * @param tUE User's type with descriptors of related to table.
	 * @param oldTable Descriptor of table created in a previous template version.
	 * @return New table descriptor.
	 */
	public fun <tUE : Any> keepTable(
		table: Table<tUE, *>,
	): Table<tUE, *>

	/**
	 * Renames or moves table without changing its schema.
	 *
	 * @param tUE User's type with descriptors of related to table.
	 * @param table Descriptor to table which should be renamed.
	 * @param newNamespace New namespace of the table. May be `null` if not changed.
	 * @param newName New namespace of the table. May be `null` if not changed.
	 * @return New table descriptor.
	 * @throws IllegalArgumentException If both [newName] and [newNamespace] are same as old or `null`.
	 */
	public fun <tUE : Any> renameTable(
		table: Table<tUE, *>,
		newNamespace: Namespace? = null,
		newName: String? = null
	): Table<tUE, *>

	/**
	 * Upgrade table's descriptors without changing its [name][Table.name] or [namespace][Table.namespacesFromModuleRoot].
	 *
	 * @param tnUE User's type with descriptors of upgraded table.
	 * @param toUE User's type with table's descriptors from the previous table version.
	 * @param oldTable Descriptor to the previous version of table from which can be got user's extension with old descriptors.
	 * @param upgrade Object with descriptor providers and factories.
	 * @return New table descriptor.
	 */
	public fun <tnUE : Any, toUE : Any> upgradeTable(
		oldTable: Table<toUE, *>,
		upgrade: TableUpgrade<tnUE, toUE>
	): Table<tnUE, toUE>


	/**
	 * Both upgrade table's schema and renames or moves table.
	 *
	 * @param tnUE User's type with descriptors of upgraded table.
	 * @param toUE User's type with table's descriptors from the previous table version.
	 * @param oldTable Descriptor of table which should be renamed and upgraded.
	 * @param newNamespace New namespace of the table. May be `null` if not changed.
	 * @param newName New namespace of the table. May be `null` if not changed.
	 * @param upgrade Object with descriptor providers and factories.
	 * @return New table descriptor.
	 * @throws IllegalArgumentException If both [newName] and [newNamespace] are same as old or `null`.
	 */
	public fun <tnUE : Any, toUE : Any> upgradeAndRenameTable(
		oldTable: Table<tnUE, toUE>,
		newNamespace: Namespace? = null,
		newName: String? = null,
		upgrade: TableUpgrade<tnUE, toUE>
	): Table<tnUE, toUE>

	/**
	 * Deletes table and all data in it.
	 *
	 * @param table Descriptor of table which should be deleted.
	 *
	 * @see UpgradeModuleScope.deleteTableAfterModuleUpgraded
	 */
	public fun deleteTable(table: Table<*, *>)

	/**
	 * Keeps table while module upgrading and deletes it after [ModuleTemplate.ModuleUpgrade.upgradeData].
	 * [Table's][table] name become free and can be used for another table.
	 *
	 * @param table Descriptor of table which should be deleted.
	 *
	 * @see UpgradeModuleScope.deleteTable
	 */
	public fun deleteTableAfterModuleUpgraded(table: Table<*, *>)


	/**
	 * Creates a temporary table that exits only while module upgrading and auto-deleted after [ModuleTemplate.ModuleUpgrade.upgradeData].
	 *
	 * @param tUE User's type for containing table-related descriptors.
	 * @param namespace Way to group tables if there is a lot.
	 * @param name Name of table for debugging.
	 * @param initializer Table-related descriptors initializer.
	 * @return Table descriptor.
	 *
	 * @see UpgradeModuleScope.deleteTableAfterModuleUpgraded
	 * @see CreateModuleScope.createTable
	 * @see CreateModuleScope.createSessionScopeTemporaryTable
	 * @see CreateModuleScope.createTransactionScopeTemporaryTable
	 */
	public fun <tUE : Any> createModuleUpgradeScopeTemporaryTable(namespace: Namespace, name: String, initializer: TableConstructor<tUE>): Table<tUE, Nothing>


	/**
	 * Upgrades [module][oldModule] located at [rootNs] with new schema template.
	 *
	 * @param rootNs Namespace where the [old module][oldModule] is located and where upgraded module will be.
	 * @param template Schema template.
	 * @return Module descriptor.
	 */
	public fun <smUE : Any> upgradeModule(oldModule: Module<*>, rootNs: Namespace, template: ModuleTemplate<smUE>): Module<smUE>
}